#!/usr/bin/env python3

import os
import sys
import time
import signal
import platform
import threading
import faulthandler
from queue import Queue

# do-or-die: coredump if a task times out
# v1.0, 2021-04-29, ed <irc.rizon.net>, MIT-licensed
# https://ocv.me/dev/?dod.py


WINDOWS = sys.platform in ["win32", "msys"]


def eprint(*a, **ka):
    ka["file"] = sys.stderr
    print(*a, **ka)


class DoOrDie(object):
    def __init__(self):
        self.q = Queue(64)
        self.deadline = None
        self.active_task = None

        self.worker_thr = threading.Thread(target=self._worker, daemon=True)
        self.worker_thr.start()

        t = threading.Thread(target=self._watchdog, daemon=True)
        t.start()

    def do(self, timeout, fun, *args, **kwargs):
        ret = Queue()
        self.q.put([timeout, fun, args, kwargs, ret])
        
        ok, ret = ret.get()
        if ok:
            return ret

        raise ret

    def shutdown(self):
        self.q.put(None)
        self.worker_thr.join()
        self.worker_thr = None

    def _worker(self):
        while True:
            task = self.q.get()
            if task is None:
                break

            timeout, fun, args, kwargs, ret_q = task

            self.active_task = task
            self.deadline = time.time() + timeout
            try:
                ret = self._exec(fun, args, kwargs)
                ret_q.put([True, ret])
            except Exception as ex:
                eprint("dod-ex: {!r}\n  {!r}\n".format(task[:-1], ex), end="")
                ret_q.put([False, ex])

            self.deadline = None

    def _exec(self, fun, args, kwargs):
        x = ["dod-exec: " + repr([fun, args, kwargs])]
        ret = fun(*args, **kwargs)
        del x[0]

    def _watchdog(self):
        while self.worker_thr:
            time.sleep(1)
            if not self.deadline:
                continue

            if time.time() >= self.deadline:
                eprint("dod-time: {!r}".format(self.active_task[:-1]))
                self._dump()
                sys.exit(1)

    def _dump(self):
        ts = time.time()

        fn = "coredump-{:.3f}.txt".format(ts)
        with open(fn, "w") as f:
            f.write("\n".join([str(x) for x in [
                platform.python_implementation(),
                sys.version_info,
                platform.system(),
                sys.platform,
                platform.python_compiler(),
                platform.version(),
                time.time(),
                self.deadline,
                repr(self.active_task)
            ]]) + "\n\n")
            faulthandler.dump_traceback(file=f)

        if WINDOWS:
            fn = "coredump-{:.3f}.dmp".format(ts)
            cw = windump(fn)
        else:
            os.kill(os.getpid(), signal.SIGABRT)


if WINDOWS:
    import ctypes
    from ctypes.wintypes import HANDLE, BOOL, DWORD, LPWSTR, LPVOID
    from ctypes import windll, WinError, get_last_error

    OpenProcess = ctypes.windll.kernel32.OpenProcess
    OpenProcess.argtypes = (DWORD, BOOL, DWORD)
    OpenProcess.restype = HANDLE

    MiniDumpWriteDump = ctypes.windll.DbgHelp.MiniDumpWriteDump
    MiniDumpWriteDump.argtypes = (HANDLE, DWORD, HANDLE, DWORD, DWORD, DWORD, DWORD)
    MiniDumpWriteDump.restype = BOOL

    CreateFile = ctypes.windll.kernel32.CreateFileW
    CreateFile.argtypes = (LPWSTR, DWORD, DWORD, LPVOID, DWORD, DWORD, HANDLE)
    CreateFile.restype = ctypes.wintypes.HANDLE

    FILE_CREATE_ALWAYS = 2
    FILE_GENERIC_RW = 0xc0000000
    FILE_ATTRIBUTE_NORMAL = 0x80
    PROCESS_ALL_ACCESS = 0x1f0fff
    COREDUMP_MODE = 2  # 0=normal 2=fullMemory

    def winfug(msg):
        raise Exception('{} ({})'.format(msg, WinError(get_last_error())))

    def windump(fn):
        pid = os.getpid()

        hproc = OpenProcess(PROCESS_ALL_ACCESS, False, pid)
        if not hproc:
            winfug('could not openprocess')

        hfile = CreateFile(fn, FILE_GENERIC_RW, 0, None, FILE_CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, None)
        if not hfile:
            winfug('could not createfile')

        res = MiniDumpWriteDump(hproc, pid, hfile, COREDUMP_MODE, 0, 0, 0)
        if not res:
            winfug('could not writedump')

        eprint('wrote coredump to [{}]'.format(fn))


def main():
    dod = DoOrDie()

    dod.do(1, print, "hey")
    
    try:
        dod.do(1, os.remove, "fgsfds")
    except Exception as ex:
        print("aa [{!r}] aa".format(ex))
    
    dod.do(1, time.sleep, 5)


if __name__ == '__main__':
    main()

